package org.fictus

class AnyRefConverter[A <: AnyRef](expected: A)(implicit control: Control, manifest: Manifest[A]) {
    def same = Matchers.same(expected)
}
class AnyConverter[A](expected: A)(implicit control: Control, manifest: Manifest[A]) {
    def equal = Matchers.equal(expected)
    def ==: = Matchers.equal(expected)
}

object Matchers {
    implicit def forAny[A](expected: A)(implicit control: Control, manifest: Manifest[A]) = new AnyConverter(expected)
    implicit def forAnyRef[A <: AnyRef](expected: A)(implicit control: Control, manifest: Manifest[A]) = new AnyRefConverter(expected)

    def equal[A](expected: A)(implicit control: Control, manifest: Manifest[A]): A = {
        control.matchers.append(Equal(expected))
        expected
    }

    def same[A <: AnyRef](expected: A)(implicit control: Control, manifest: Manifest[A]): A = {
        control.matchers.append(Same(expected))
        expected
    }

    // TODO the syntax for this is problematic.
    // using * with no type parameter leads to the type being Object
    // using *[type] works, but is ugly
    // this is relevant http://stackoverflow.com/questions/9118264/multiple-parameter-closure-argument-type-not-inferred
    def *[A](implicit control: Control, manifest: Manifest[A]): A = any

    def any[A](implicit control: Control, manifest: Manifest[A]): A = {
        val default = Types.defaultValue(manifest.erasure).asInstanceOf[A]
        control.matchers.append(Anything[A](default))
        default
    }
}

trait Matcher {
    val underlyingClass: Class[_]
    val expected: Any
    def matches(actual: Any): Boolean
}

case class Equal[A: Manifest](expected: A) extends Matcher {
    override val underlyingClass = manifest.erasure
    override def matches(actual: Any) = expected == actual
}

case class Same[A <: AnyRef: Manifest](expected: A) extends Matcher {
    override val underlyingClass = manifest.erasure
    override def matches(actual: Any) = actual match {
        case actual: AnyRef => expected eq actual
        case _              => false
    }
}

case class Anything[A: Manifest](expected: A) extends Matcher {
    override val underlyingClass = manifest.erasure
    override def matches(actual: Any) = true
}
